/**
* Title: 		 JavaTester
* Description:	 Class testing some Java 5 features
* Copyright:     Copyright (c) 2007-2014
* Company:       Morpheus Quality Technologies
* Creation date: 04-10-2009
* Last modified: 30-12-2014
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.tutorial.jfive ;

import java.util.* ;

import static java.lang.Math.* ;	// Necessary to test the testImportStatic() method

/**
* The methods offered by this class are used by the <code>SimpleClient</code> to test the new features
* introduced by Java5.<br/><br/>
* 
* To invoke a method of the <code>JavaTester</code> class, run the <code>SimpleClient</code> using
* the <code>-j</code> option followed by a number between 0 and 7 (included).
*/
public class JavaTester {

	/** 
	* Definition of an anonymous <code>Enum</code> class listing the basic "coding related" Java5 features.	
	* See the <code>testEnumTypes</code> method for details.
	* 
	* @see #testEnumTypes()
	*/
	public 					enum 		My_enum
	{ GENERICS, AUTOBOXING, FORLOOP, ENUM, CONSTANTS, PRINTF, SCANNER, VARARGS } ;

  	/** Beginning of each logging message produced by this class.											*/
	protected static		String		leadingName 		= "JavaTester - " ;

    /**
	* Test some features introduced by the Java 5 version: the feature to be tested is selected by
	* the input <code>code</code> integer parameter:<br/><br/>
	*
	* <strong>0</strong>: test generics types.<br/>
	* <strong>1</strong>: test auto-boxing and auto-unboxing for generics types.<br/>
	* <strong>2</strong>: test enhanced for loop.<br/>
	* <strong>3</strong>: test the new <code>enum</code> object types.<br/>
	* <strong>4</strong>: test the import static keyword.<br/>
	* <strong>5</strong>: test formatted output.<br/>
	* <strong>6</strong>: test formatted input.<br/>
	* <strong>7</strong>: test the varargs functionality.<br/>
	*
	* @param code the integer index used to select the Java 5 feature
    */
    public void testJava5(int code) {

		switch(code) {
			case 	0	: testGenerics() ;										break ;
			case 	1	: testAutoboxing() ;									break ;
			case 	2	: testForLoop() ;										break ;
			case 	3	: testEnumTypes() ;										break ;
			case 	4	: testImportStatic() ;									break ;
			case 	5	: testFormattedOutput() ;								break ;
			case 	6	: testFormattedInput() ;								break ;
			case 	7	: testArguments('a', code, this) ;						break ;
			default		: System.out.println("Please specify a number between 0 and 7") ;
		}
	}

// ************************************************************************************************************
// ********************************			BASIC JAVA 5				***************************************
// ************************************************************************************************************

    /**
	* Test the "generic" types, i.e. the ability to specify the object type when using collections (or similar),
	* using a syntax like: <code>ArrayList&lt;Integer&gt; list =  new ArrayList&lt;Integer&gt;()</code>.
	* The advantage, from the coding point of view, it's to avoid the explicit casting when getting
	* objects stored into a list or collection (there are, of course, also some improvements from the
	* performances point of view).<br/><br/>
	* 
	* This feature is coupled to the auto-boxing and auto-unboxing of primitive types, showed by the
	* <code>testAutoboxing()</code> method.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-j 0</code> option.<br/>
	* 
	* @see #testAutoboxing()
    */
    public void testGenerics() {

		// Initializing
		int max = 10 ;
		System.out.println(leadingName + "Testing generics types\n") ;

		// Filling array list
		ArrayList<Integer> list = new ArrayList<Integer>() ;
		for (int i = 0 ; i < max ; i++) list.add(i, new Integer(10 + i)) ;

		// Reading array list
		int total = 0 ;
		for (int i = 0 ; i < max ; i++) {
			// No casting! In Java 1.4 this was: value = ( (Integer) list.get(i) ).intValue() ;
			int value = list.get(i).intValue() ;
			System.out.println(leadingName + "Read value: " + value) ;
			total += value ;
		}
		System.out.println(leadingName + "Total: " + total) ;
	}

    /**
	* Test the auto-boxing and auto-unboxing of primitive types, i.e. the possibility to avoid parsing methods
	* when setting/getting primitives values from the corresponding objects (e.g. from int to Integer).<br/>
	* Notice that these features (auto-boxing and auto-unboxing) works only if we define the collection object
	* using the "generic" syntax (see also the <code>testGenerics()</code> method).<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-j 1</code> option.<br/>
	* 
	* @see #testGenerics()
    */
    public void testAutoboxing() {

		// Initializing
		int max = 10 ;
		System.out.println(leadingName + "Testing auto-boxing and auto-unboxing of primitive types\n") ;

		// Filling array list
		ArrayList<Integer> list =  new ArrayList<Integer>() ;
		// Try to use the line below to verify that auto-boxing and auto-unboxing do not work on raw types 
		// ArrayList list =  new ArrayList() ;
		
		// Auto-boxing: In Java 1.4 this was list.add(0, new Integer(10 + i))	(see testGenerics method)
		for (int i = 0 ; i < max ; i++) list.add(i, 10 + i) ; 	

		// Reading array list
		int total = 0 ;
		for (int i = 0 ; i < max ; i++) {
			// Auto-unboxing: In Java 1.4 this was list.get(i).intValue() ;		(see testGenerics method)
			int value = list.get(i) ;							
			System.out.println(leadingName + "Read value: " + value) ;
			total += value ;
		}
		System.out.println(leadingName + "Total: " + total) ;
	}

    /**
	* Testing enhanced for loop. Now loops can be written as:<br/><br/>
	*
	* <code>for (int value : list) { ...</code><br/>
	* <code>for (Integer value : list) { ...</code><br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-j 2</code> option.<br/>
    */
    @SuppressWarnings({"rawtypes", "unchecked"})
    public void testForLoop() {

		// Initializing
		int max 	= 10 ;
		int total 	= 0 ;
		System.out.println(leadingName + "Testing enhanced for loop\n") ;

		// OLD SYSTEM (Java4)
		ArrayList old =  new ArrayList() ;
		for (int i = 0 ; i < max ; i++) old.add(i, new Integer(10 + i)) ;
		// Old alternative syntax (implicit increment condition)
		for (Iterator i = old.iterator() ; i.hasNext() ; ) {
			Integer value = ((Integer) i.next()).intValue() ;
			System.out.println(leadingName + "Read value (old system): " + value) ;
			total += value ;
		}
		System.out.println(leadingName + "Total: " + total +  "\n") ;

		// NEW SYSTEM (Java5)
		total = 0 ;
		ArrayList<Integer> list =  new ArrayList<Integer>() ;
		for (int i = 0 ; i < max ; i++) list.add(i, 10 + i) ;
		// Enhanced For loop example
		for (int value : list) {
			System.out.println(leadingName + "Read value (new system): " + value) ;
			total += value ;
		}
		System.out.println(leadingName + "Total: " + total +  "\n") ;
	}

    /**
	* Testing <code>enum</code> types.<br/>
	* Notice that all enums implicitly extend <code>java.lang.Enum</code>. Hence, since Java does not support
	* multiple inheritance, an <code>enum</code> cannot extend anything else. Moreover it's not possible
	* to extend an <code>enum</code> class to obtain a child class inheriting some already defined
	* enumeration: if you need very similar <code>enum</code> classes you must rely on the old
	* "copy and paste code" mechanism.<br/><br/>
	* 
	* This test is divided into two sections:<br/><br/>
	* 
	* The first part read all the values from the <code>My_enum</code> object (defined as class field) and
	* print the enum keys on stdout, printing a special message for one of the enum's keys.<br/><br/>
	* 
	* The second part shows an example of retrieving formatted data from an external <code>enum</code> class,
	* in this case the <code>adriani.tutorial.jfive.Library</code> class.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-j 3</code> option.<br/>
	* 
	* @see Library
    */
    public void testEnumTypes() {

		// Testing basic enum usage
		System.out.println(leadingName + "Testing enum types (part 1)\n") ;
		for (My_enum j : My_enum.values()) {
			System.out.println(leadingName + "Enum value: " + j) ;
			if (j == My_enum.ENUM) System.out.println(leadingName + "It's me! I am the My_enum.ENUM type!") ;
		}

		// Testing advanced enum usage
		System.out.println("\n" + leadingName + "Testing enum types (part 2)\n") ;
		Library.main(null) ;
	}

    /**
	* Shows the capability to access constant static class fields without qualification.<br/>
	* See the <code>import static java.lang.Math.*</code> statement in the class header.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-j 4</code> option.<br/>
    */
    public void testImportStatic() {

		System.out.println(leadingName + "Testing the import static keyword\n") ;
		System.out.println(leadingName + "Math.E  = " + E) ;
		System.out.println(leadingName + "Math.PI = " + PI + "\n") ;
	}

    /**
	* Implement the <code>printf</code> funtion using the C++ syntax.<br/>
	* The generic syntax for the formatting pattern is the following:<br/><br/>
	* 
	* <code>%[argument_index$][flags][width][.precision]conversion</code><br/><br/>
	* 
	* For details see the "Format String Syntax" section in the <code>java.util.Formatter</code>
	* API documentation page.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-j 5</code> option.<br/>
	* 
	* @see java.util.Formatter
    */
    public void testFormattedOutput() {
    	
    	// How to read formatting instructions:
    	// - All formatting instructions starts with the symbol "%"  
    	// - The symbol "-" is an option flag used to state that the result will be left-justified.
    	// - The symbol "s" means "format as a string"
    	// - The symbol "d" means "formate as a decimal number"
    	// - The symbol "%n" means "insert a new line separator"

		System.out.println(leadingName + "Testing the formatted C++ like output syntax\n") ;
		System.out.printf("%-20s %s%n"	, "Author"			, "Books"	) ;
		System.out.printf("%-20s %5d%n"	, "D. Adams"		, 42		) ;
		System.out.printf("%-20s %5d%n"	, "J.R.R. Tolkien"	, 3			) ;
		System.out.println() ;
	}

    /**
	* Test the <code>java.util.Scanner</code> class useful to automatically parse values from
	* <code>stdin</code>.<br/><br/>
	* 
	* For details see the <code>java.util.Scanner</code> API documentation.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-j 6</code> option.<br/>
	* 
	* @see java.util.Scanner
    */
    public void testFormattedInput() {

		// Reading data
		System.out.println(	leadingName + "Testing the formated input scanner") ;
		System.out.println	("Please insert a string and a number.") ;
		System.out.print	("You can insert these input on the same line, space separated, ") ;
		System.out.print	("or on two different lines\n\n") ;
		Scanner s = new Scanner(System.in) ;
		// Using the new Scanner methods
		String param 	= s.next() ;
		int value 		= s.nextInt() ;
		s.close() ;

		// Printing data
		System.out.println() ;
		System.out.println(leadingName + "param = " + param) ;
		System.out.println(leadingName + "value = " + value) ;
	}

    /**
	* Test the new varargs functionality, allowing to pass a variable number of arguments, of different
	* types, without the need to wrap input data inside an array or a collection. This example show
	* the result of following call:<br/><br/>
	*
	* <code>testArguments('a', code, this)</code><br/><br/>
	*
	* To test this method run the <code>SimpleClient</code> using the <code>-j 7</code> option.<br/>
	* 
	* @param args any sequence of Java objects or Java primitives, in any order
    */
    public void testArguments(Object ... args) {

		// Header
		System.out.println(leadingName + "Testing the varargs list feature.") ;
		System.out.println("This method has been implementad using this syntax:\n\n" ) ; 
		System.out.println("\tpublic void testArguments(Object ... args) {}\n\n" ) ;
		System.out.println("invoked in this way:\n\n" ) ;		
		System.out.println("\ttestArguments('a', code, this)\n\n" ) ;
		System.out.println("and providing this result:\n\n" ) ;

		// Printing metadata
		System.out.println(leadingName + "Arguments array length : "	+ args.length) ;
		System.out.println(leadingName + "Arguments class        : "	+ args.getClass()) ;
		System.out.println(leadingName + "Arguments instance     : "	+ args) ;
		System.out.println() ;

		// Printing arguments
		System.out.println("Listing all arguments:\n" ) ;
		for (int i = 0 ; i < args.length ; i++) {
			System.out.print("Param " + i + " = " + args[i] + "\t") ;
			System.out.println("(" + args[i].getClass() + ")") ;
		}
    }

} // End of class