/**
* Title: 		 Library
* Description: 	 Testing Java enumeration type
* Copyright:     Copyright (c) 2008-2013
* Company:       Morpheus Quality Technologies
* Creation date: 04-10-2009
* Last modified: 30-12-2014
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.tutorial.jfive ;

/** An example of enum types defined in a custom class: this example show a simple enum list depicting a set 
* of key-value entry (e.g. this could describe the number of files with related overall dimensions).<br/><br/>
*
* Usage: <code>adriani.tutorial.jfive.Library</code>.<br/><br/>
*
* The advantage to use the <code>enum</code> type, compared to basic Java sets, is the possibility to add
* inner methods to the enumeration, even a <code>main()</code> method!<br/>
* In other words, the <code>enum</code> type is a wrapper for static collections, offering a smart syntax
* which allows to save coding time (e.g. easy definitions of LABELS and related CONSTANTS).
*/
public enum Library {

	/** Enum definition (MANDATORY).																		*/
	ROOT		(1	, 3.686e3	)	,
	CLIENT		(4	, 53.996e3	)	,
	DB			(7	, 102.925e3	)	,
	ECLIPSE		(11	, 186.533e3	)	,
	MESSAGGES	(7	, 65536		)	,
	RESOURCES	(12	, 19.927e3	)	,
	SERVICE		(7	, 74.485e3	)	,
	UTIL		(7	, 55.771e3	)	;

	/** The first value bound to each enum entry.															*/
	private final 		int 	first ;

	/** The second value bound to each enum entry.															*/
	private final 		double 	second ;
	
	/** 
	* Constructor: this is mandatory in order to define the structure of the enum type.
	* Without the constructor the compiler could not understand how the class fields (belong
	* to the enumeration) must be initialized (some enum entry could be defined here).
	*/
	Library(int f, double s) {

		this.first 	= f ;
		this.second = s ;
	}

// ************************************************************************************************************
// **********************************			ENUM METHODS			***************************************
// ************************************************************************************************************

	/** Accessor. 																							*/
	private int first() 							{			return first ; 								}

	/** Accessor. 																							*/
	private double second() 						{ 			return second ;	 							}

	/** Conversion utility.																					*/
	private String getKiloBytes() 					{          	return getKiloBytes(second) ;				}

	/** Conversion utility.																					*/
	private String getKiloBytes(double value) {

		String 	label = new String("" + value/1024) ;
		int 	index = label.indexOf('.') ;
		if (index > 0) label = label.substring(0, index + 2) + " KB" ;
		return label ;
	}

	/** Formatter utility.																					*/
	private String getFormattedData() {

		String 	blanks 	= "" ;
		int 	max 	= 10 - name().length() ;
		for (int i = 0 ; i < max ; i++) blanks += " " ;
		String output = blanks + " = " + getKiloBytes() + "   (" + first + " keys)" ;
		return output ;
	}

// ************************************************************************************************************
// **********************************			MAIN METHOD 			***************************************
// ************************************************************************************************************

	/**
	* Main method.
	*/
	public static void main(String[] args) {

		// Initiliazing
		int 	classes = 0 ;
		double 	total 	= 0 ;

		// Looping through all enum types
		for ( Library p : Library.values() ) {
			total 	+= p.second() ;
			classes += p.first() ;
			System.out.println("Enum entry " + p + p.getFormattedData()) ;
		}

		// Printing result
		System.out.println() ;
		System.out.println("Total keys: " + classes) ;
		// Remark: in this case we could call the "getKiloBytes(total)" method on ANY enum types
		// (i.e. ROOT is just an arbitrary choice)
		System.out.println("Total values: " + ROOT.getKiloBytes(total) + " (" + total + " bytes)") ;
	}

} // End of enum