/**
* Title: 		 SimpleClient
* Description:	 Simple application for testing Java 5 features 
* Copyright:     Copyright (c) 2013-2015
* Company:       Morpheus Quality Technologies
* Creation date: 04-10-2009
* Last modified: 23-02-2015
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.tutorial.jfive ;

import adriani.tutorial.jfive.generic.GenericTester;
import adriani.tutorial.jfive.generic.WildcardTester;

/**
* A simple command line class offering a <code>main()</code> method to allow invocation of some 
* Java 5 features directly from the command line.<br/>
* If all the necessary environment variables are correctly defined it should be possible to run
* this application by writing:<br/><br/>
*
* <code>java -cp %CLASSPATH% adriani.tutorial.jfive.SimpleClient</code><br/><br/>
*
* For further details run <code>SimpleClient -h</code>.
*/
public class SimpleClient {
	
	/** Static final value storing the default run option.													*/
	public static final		String[]		defPar 				= new String[] { "-h",  ""  } ;
	//public static final	String[]		defPar 				= new String[] { "-j",  "0" } ;

	/** Static final value stored as class field to improve performances.									*/
	public static final		String			EMPTY 				= "" ;
	/** Static final value stored as class field to improve performances.									*/
	public static final		String			BLANK 				= " " ;
	/** Static final value stored as class field to improve performances.									*/
	public static final	 	String			ZERO 				= "0" ;
	/** Static final value stored as class field to improve performances.									*/
	public static final 	StringBuffer	NULL 				= new StringBuffer("NULL") ;

  	/** Beginning of each logging message produced by this class.											*/
	protected static 		String			leadingName 		= "Simple Client - " ;

    /**
	* Main method used to call the application from the command line, forwarding action to the
	* <code>run(String[] args)</code> method.<br/><br/>
	*
	* Default behavior (no option at all) will launch the simple client running the first method
	* of the Java5 test section, which should correspond to the option "-j 0".<br/><br/> 
	*
	* @param args the standard input string array, usually read from STDIN
	* @see run(String[] args)
    */
    public static void main(String[] args) {

		// Creating the client
		SimpleClient client = new SimpleClient() ;

		// Specify the default method number here
		client.run(args, SimpleClient.defPar) ;

		// Comment next line if necessary (e.g. if you have other threads running)
		System.exit(0) ;
	}

    /**
	* Check the number of input parameters: if the client has been invoked with some parameters action is 
	* forwarded to the <code>manageOptions</code> method. Otherwise, if the input argument is missing, 
	* the <code>manageOptions</code> method is invoked using the default setting (which should be "-j 0").
	*
	* @param args the standard input string array, usually read from STDIN
	* @param defPar the test method to be invoked (if 0 we'll use the default one, usually the first method)
    */
    public void run(String[] args, String[] defaultParams) {

    	// 1. Logging number of input parameters
    	if (Settings.VERBOSE) {
    		String runInfo = (args != null) ? (args.length + " args") : "no args" ;
    		System.out.println(leadingName + "Running with " + runInfo) ;
    		// System.out.println(leadingName + "Using java version " + System.getProperty("java.version")) ;
    	}

		// 2. Eventually managing options
		if ( (args != null) && (args.length > 0) && (args[0].startsWith("-")) ) {
			if (args[0].length() > 1) manageOptions(args) ;
			else System.out.println("\nPlease specify a valid option, or use -h to get some help") ;

		// 3. Otherwise we run the default Java 5 test (see method)
		} else {
			if (Settings.VERBOSE) System.out.println(leadingName + "Using default launch configuration (use -h to get all launch configurations)") ;
			manageOptions(defaultParams) ;	
		}

		// 4. Ending
		if (Settings.VERBOSE) System.out.println("\n" + leadingName + "ending run() method ...") ;
	}

// *********************************************************************************************************
// **********************************		MANAGING OPTIONS		****************************************
// *********************************************************************************************************

    /**
	* Manage all the possible options available from the command line.
	* The available test offered by the <code>SimpleClient</code> are executed by calling:<br/><br/>
	*
	* <code>java -cp %CLASSPATH% adriani.tutorial.jfive.SimpleClient [OPT] n</code><br/><br/>
	*
	* Where the <code>OPT</code> parameter should be one of the following:<br/><br/>
	*
	* -h Print some help on line about the usage if this application.<br/>
	* -j Executes one of the Java 5 test features: see the <code>JavaTester</code>.<br/>
	* -g Executes one of the Java 5 generic test: see the <code>GenericTester</code>.<br/>
	* -w Executes one of the Java 5 generic wildcards test: see the <code>WildcardTester</code>.<br/>
	* -u Executes one specific unit test: see the <code>UnitTester</code>.<br/>
	*
	* @param args the input string array
    */
    public void manageOptions(String[] args) {

		// Converting String[] array into StringBuffer[] array
		String 			option = args[0].substring(0,2) ;
		StringBuffer[] 	buffer = new StringBuffer[args.length] ;
		for (int i = 0 ; i < args.length ; i++) {
			buffer[i] = new StringBuffer(args[i]) ;		
		}
		// Printing help screen
		if (option.equalsIgnoreCase("-h")) {
			printHelp() ;
		}
		// Testing a Java5 feature
		else if (option.equalsIgnoreCase("-j")) {
			int 		code 	= getIntValue(buffer, 1) ;
			JavaTester 	tester 	= new JavaTester() ;
			tester.testJava5(code) ;
		}
		// Testing a Generic feature
		else if (option.equalsIgnoreCase("-g")) {
			int 			code 	= getIntValue(buffer, 1) ;
			GenericTester 	tester 	= new GenericTester() ;
			tester.testGeneric(code) ;
		}
		// Testing a Generic Wildcard feature
		else if (option.equalsIgnoreCase("-w")) {
			int 			code 	= getIntValue(buffer, 1) ;
			WildcardTester 	tester 	= new WildcardTester() ;
			tester.testWildcards(code) ;
		}
		// Handling invalid options
		else {
			System.out.println("\nInvalid option: run using '-h' to list all available options\n") ;
		}
	}

// *********************************************************************************************************
// **********************************		UTILITIES			********************************************
// *********************************************************************************************************

	/**
	* Return the integer number specified after a general "-option" keyword, for example -j or -g.
	*
	* @param string the input string array, converted into a <code>String</code> object
	* @param index the index of the element to be extracted
	* @return the integer number corresponding to the index-th element of the array
    */
    public int getIntValue(String[] string, int index) {

		int code = 0 ;
		if (string.length <= index) return code ;
		try 				{ code = Integer.parseInt( string[index].trim() ) ;					}
		catch(Exception e) 	{ code = 0 ;														}
		return code ;
	}

    /**
	* Return the integer number specified after a general "-option" keyword, for example -j or -g.
	*
	* @param buffer the input string array, converted into a <code>StringBuffer</code> object
	* @param index the index of the element to be extracted
	* @return the integer number corresponding to the index-th element of the array
    */
    public int getIntValue(StringBuffer[] buffer, int index) {

		int code = 0 ;
		if (buffer.length <= index) return code ;
		try 				{ code = Integer.parseInt( buffer[index].toString().trim() ) ;		}
		catch(Exception e) 	{ code = 0 ;														}
		return code ;
	}
    
    /**
	* Adjust the content of the input String in a friendly printable format. 
	*
	* @param array the input <code>String</code> array object
	* @return a String containing the array content
    */
    public String formatArray(String[] array) {

    	String result = "" ;
    	for (String element : array) result += element.toString() + BLANK ;
    	return result ;
	}    

    /**
	* Print some help about the <code>SimpleClient</code> command line options on the STDOUT.
	*
	* @return the formatted string message printed on STDOUT
    */
    public StringBuffer printHelp() {

		// Preparing text message
		StringBuffer message = new StringBuffer() ;
		message.append("\n") ;
		message.append("jFive - Testing Java 5 features and unit tests on the command line, ") ;
		message.append("version " + Settings. VERSION + "\n") ;
		message.append("\n") ;
		message.append("Usage: java -jar jfive.jar [options]\n") ;
		message.append("or     java -cp %CLASSPATH% adriani.tutorial.jfive.SimpleClient [options]\n") ;
		message.append("\n") ;
		message.append("DEFAULT BEHAVIOUR\nIf no option is specified, the client is launched using the option " ) ;
		message.append(formatArray(SimpleClient.defPar) +  "\n") ;
		message.append("\n") ;
		message.append("OPTIONS\n") ;
		message.append("-h          print this help screen\n") ;
		message.append("-j n        executes the j-th method from the JavaTester class\n") ;		
		message.append("-g n        executes the g-th method from the GenericTester class\n") ;		
		message.append("-w n        executes the w-th method from the WildcardTester class\n") ;		
		message.append("\n") ;

		// Producing output
		System.out.println(message) ;
		return message ;
	}

} // End of class