/**
* Title: 		 JavaTester
* Description:	 Class testing some Java 5 "Generic" features
* Copyright:     Copyright (c) 2013-2015
* Company:       Morpheus Quality Technologies
* Creation date: 30-12-2014
* Last modified: 10-01-2015
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.tutorial.jfive.generic ;

import java.util.Collections;
import java.util.List ;

/**
* Invokes many different methods on some dummy objects, useful to test the main Java 5 
* Generic features. Results obtained from this test sections could be summarized as follows:<br/><br/>
* 
* - Generic CLASSES are identified writing the Generic type AFTER the class: <code>MyClass&lt;T&gt;</code><br/>
* - Generic METHODS are identified writing the Generic input type BEFORE the method's name: <code>&lt;T&gt;myMethod()</code><br/><br/>
*
* To invoke a method of the <code>JavaTester</code> class, run the <code>SimpleClient</code> using
* the <code>-g</code> option followed by a number between 0 and 6 (included). See below for the list of
* the available test methods.
* 
* @see <a href="http://docs.oracle.com/javase/tutorial/java/generics/index.html">Java SE tutorial</a>  
*/
public class GenericTester {

  	/** Beginning of each logging message produced by this class.											*/
	protected static		String		leadingName 		= "GenericTester - " ;

    /**
	* To invoke a single test method specify one these input <code>code</code> integer parameters:<br/><br/>
	*
	* <strong>0</strong>: test Generic creations versus raw types.<br/>
	* <strong>1</strong>: test complex types (i.e. a "nested" Generic).<br/>
	* <strong>2</strong>: test Generic methods signature (taking a Generic input).<br/>
	* <strong>3</strong>: test Generic Bounded Type signature (input type constraints).<br/>
	* <strong>4</strong>: test Generic Multiple Bounds (input type multiple constraints).<br/>
	* <strong>5</strong>: test Generic methods signature (taking a Generic input).<br/>
	* <strong>6</strong>: test Generic inheritance (extending a Generic type).<br/>
	* <strong>7</strong>: test advanced type inference.<br/>
	*
	* @param code the integer index used to select the Java 5 Generic feature
    */
    public void testGeneric(int code) {
    	
    	switch(code) {
			case 	0	: creationTest() ;										break ;
			case 	1	: testComplexTypes() ;									break ;
			case 	2	: testGenericMethod() ;									break ;
			case 	3	: testBoundedType() ;									break ;
			case 	4	: testMultipleTypes() ;									break ;
			case 	5	: testGenericAlgorithms() ;								break ;
			case 	6	: testInheritance() ;									break ;
			case 	7	: testTypeInference() ;								break ;
			default		: System.out.println("Please specify a number between 0 and 7") ;
		}
	}

    /**
	* Test how to create a Generic type and a raw type, using the dummy <code>MyGeneric</code> class
	* (included in this project).
	* The best way to see the differences between the two approaches is at compile-time:  
	* look at the implementation of this method and try to compile it under different conditions 
	* (i.e comment/uncomment some lines or just read IDE warnings).<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-g 0</code> option.<br/>
	* 
	* @see MyGeneric
    */
    @SuppressWarnings({"rawtypes","unchecked"})
    public void creationTest() {
    
    	// 1. This is a generic type invocation: the type can be considered a part of the class's name.
    	MyGeneric<String> 	obj_1 = new MyGeneric<String>() ;
    	// Creating a "raw type", i.e. a Generic object without type a type argument.
    	MyGeneric			obj_2 = new MyGeneric() ;
    	// Shorter syntax (available since Java SE 7): using the "diamond"
    	// MyGeneric<String> obj_2 = new MyGeneric<>() ;
    	
    	// 2. Showing differences among raw type and generic type:
    	obj_1.set("I am a type argument (a String)!"	) ;
    	obj_2.set("I am a normal argument (an Object)!"	) ;
    	System.out.println("\n" + leadingName + "Object 1 trimmed: " + obj_1.get().trim()) ;
    	// This cannot work, since the raw type stores an Object, not a String!
    	// System.out.println("\n" + leadingName + "Object 1 trimmed: " + obj_2.get().trim()) ;
    	
    	// 3. Providing some feedback
    	System.out.println("\nMethod creationTest() succesfull ended...\n") ;
    }    	
    	
    /**
	* Test how to create a complex type, i.e. a Generic type "wrapping" another
	* Generic type. In this example the <code>Pair</code> type is created by including
	* (as "inner" type") the <code>MyGeneric</code> type.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-g 1</code> option.<br/>
	* 
	* @see Pair
	* @see MyGeneric 
    */
    @SuppressWarnings("unused")
    public void testComplexTypes() {
    	
    	// 1. Creating a simple Generic type
    	MyGeneric<String> obj_1 = new MyGeneric<String>() ;
    	
    	// 2. An example of complex Generic (i.e. a "nested" type)
    	Pair< Integer, MyGeneric<String> > pair = new Pair< Integer, MyGeneric<String> >(1, obj_1) ;
    	
    	// 3. Providing some feedback
    	System.out.println("\nMethod testComplexTypes() succesfull ended...\n") ;
    }    	

    /**
	* Test the syntax for defining Generic methods. This method shows as following syntaxes are both allowed:<br/><br/>
	* 
	* <code>Pair.&lt;Integer, String&gt;compare(pair_1, pair_2) ;</code><br/>
	* <code>Pair.compare(pair_1, pair_2) ;</code><br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-g 2</code> option.<br/>
	* 
	* @see Pair 
    */
    public void testGenericMethod() {
    	
    	// 1. Creating objects (without using the "diamond")
    	Pair<Integer, String> pair_1 = new Pair<Integer, String>(1, "Hello") ; 
    	Pair<Integer, String> pair_2 = new Pair<Integer, String>(1, "Hello") ;
    	
    	// 2.1 Invoking method using long syntax: specifies the "type witness", i.e. "<Integer, String>"
    	boolean compare_1 = Pair.<Integer, String>compare(pair_1, pair_2) ;
    	// 2.2 Invoking method using a shorter syntax (TYPE INFERENCE): saves typing (no "type witness")
    	boolean compare_2 = Pair.compare(pair_1, pair_2) ;
    	
    	// 3. Providing some feedback
    	System.out.print	("\nRunning testGenericMethod() method: ") ;
    	System.out.println	("this method shows how to invoke a generic method using two syntaxes:\n\n") ;
    	System.out.println	("Syntax 1: Pair.<Integer, String>compare(pair_1, pair_2)") ;
    	System.out.println	("Syntax 2: Pair.compare(pair_1, pair_2) ;") ;
    	
    	// 4. Printing results
    	System.out.println("\nShowing results of these two invocations:\n") ;
    	System.out.println("Comparing pairs (1): "	+ compare_1) ;
    	System.out.println("Comparing pairs (2): "	+ compare_2) ;
    	System.out.println() ;
    }
    
    /**
	* Test the syntax for defining <strong>Bounded Type</strong> Generic methods, which allows to set some constraints 
	* on a Generic method. In this example we invoke the method:<br/><br/>
	* 
	* <code>public &lt;T extends CharSequence&gt; void inspect(T t){}</code><br/><br/>
	* 
	* hence the method can be invoked only using parameters which extends the <code>CharSequence</code> class.
	* To test this method run the <code>SimpleClient</code> using the <code>-g 3</code> option.<br/>
	* 
	* @see Pair  
    */
    public void testBoundedType() {
    	
    	// 1. Defining a simple Pair type
    	Pair<Integer, String> pair_1 = new Pair<Integer, String>(1, "Hi there...") ;
    	
    	// 2. Providing some feedback
    	System.out.println("\nTesting method testBoundedType():\n") ;
    	
    	// 3. Testing BOUNDED TYPE PARAMETERS (StringBuffer inherits the CharSequence superclass)
    	pair_1.inspect( new StringBuffer("42") ) ;
    	// This cannot compile, since "inspect()" accepts only CharSequence subclasses 
    	// pair_1.inspect(42);
    }
    
    /**
	* Test the syntax for defining <strong>Multiple Bounds</strong>, which allows to set more strict constraints on 
	* Generic methods. In this example we invoke the method:<br/><br/>
	* 
	* <code>public &lt;T extends ParentOne &amp; ParentTwo&gt; void multiple(T t){</code><br/><br/>
	* 
	* hence the method can be invoked only using parameters which extends both the <code>ParentOne</code> class 
	* and the <code>ParentTwo</code> interface (all these objects are defined as inner classes).<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-g 4</code> option.<br/>
	* 
	* @see Pair
	* @see ParentOne
	* @see TestMultiple 
    */
    @SuppressWarnings("unused")
    public void testMultipleTypes() {
    	
    	// 1. Defining a simple Pair type
    	Pair<Integer, String> pair_1 = new Pair<Integer, String>(1, "Hello World!") ;
    	
    	// 2. Providing some feedback
    	System.out.println("\nTesting method testMultipleTypes():\n") ;
    	
    	// 3. Testing MULTIPLE BOUNDS (see inner classes definition) 
    	TestMultiple multiple 	= new TestMultiple() ;
    	ParentOne 	ParentOne 	= new ParentOne() ;
    	pair_1.multiple(multiple) ;
    	// This cannot compile, since "ParentOne" does not extend both ParentOne & ParentTwo
    	// pair_1.multiple(ParentOne) ;
    }
    
    /**
    * Example of implementation of "Generic algorithms" using the <strong>bounded type</strong> 
    * parameters.<br/> 
    * This test shows how to implement an algorithm which exploits specific parameter's methods, 
    * as long as the input type parameters share a common interface API. In this example the input types 
    * must extend the <code>Comparable</code> interface: this allows us to call the <code>compareTo()</code>
    * method on the input arguments.<br/><br/>
    * 
    * <strong>Remark</strong>: notice the different between writing a method as "T extends CharSequence" (example 3) 
    * and as "T extends Comparable&lt;T&gt;" (this example). The first syntax is allowed since CharSequence
    * is a raw type, the second syntax is mandatory since Comparable is not a raw type.<br/><br/>
    * 
	* To test this method run the <code>SimpleClient</code> using the <code>-g 5</code> option.<br/>
	* 
	* @see Pair   
    */
    public void testGenericAlgorithms() {
    	
    	// 1. Providing some feedback
    	System.out.println("\nTesting method testGenericAlgorithms():\n") ;
    	System.out.println("This method invokes the Pair.compare() static method using two types of array:") ;
    	System.out.println("first a String[] array, then an Integer[] array.") ;
    	System.out.println("The method works in both cases since it's defined using this syntax:\n") ;
    	System.out.println("\tpublic static <T extends Comparable<T>> int compare(T[] anArray, T elem) {}\n") ;
    	System.out.println("(the method just counts the number of array elements bigger than the second argument)") ;
    	System.out.println() ;
    	
    	// 2. Calling the compare() method using a String array
    	String[] string_arr = new String[] {"a", "b", "c", "d" , "e", "f"} ;
    	System.out.println ("Compare result (String)  : " + Pair.compare(string_arr, "a") ) ;
    	
    	// 3. Calling the compare() method using an Integer array
    	Integer[] int_arr = new Integer[] {1, 2, 3, 4 , 5, 6} ;
    	System.out.println ("Compare result (Integer) : " + Pair.compare(int_arr, 6) ) ;
    	System.out.println() ;
    }       
    
    /**
    * A common mistake when managing Generic inheritance is the following: even if "Integer" extends
    * "Number", this does not mean that "MyClass&lt;Integer&gt;" extends "MyClass&lt;Number&gt;".  
    * To manage this types of inheritance we must rely on "Generic wildcards", which are better explained 
    * in the <code>WildcardsTester</code> class.</br></br>
    * 
    * The only simple inheritance mechanism (regarding Generic) applies when both conditions are verified:<br/>
    * <ol>
    * <li>The 1<sup>st</sup> Generic class extend the 2<sup>nd</sup> Generic class (e.g. List extends Collection)</li>
    * <li>The type parameter is the same for both classes</li>
    * </ol>
    * 
    * Example: List&lt;MyType&gt; extends Collection&lt;MyType&gt;.<br/><br/>
    * 
	* This method shows some examples of the inheritance issue, exploiting the <code>SubType</code> interface 
	* (this object is defined as inner classes).<br/><br/>
    *  
	* To test this method run the <code>SimpleClient</code> using the <code>-g 6</code> option.<br/><br/>
	* 
	* @see SubType
	* @see MyGeneric
    */
    @SuppressWarnings("unused")    
    public void testInheritance() {
    	
    	// 1. Providing some feedback
    	System.out.println("\nTesting inheritance:\n") ;
    	System.out.println("\tonly class SubType<K,V> extends MyGeneric<K>\n") ;
    	System.out.println("Since they share the first type (i.e. K). This means that:\n") ;
    	System.out.println("\tSubType<Number, Integer> extends MyGeneric<Number> ? true") ;
    	System.out.println("\tSubType<Integer, String> extends MyGeneric<Number> ? false") ;
    	System.out.println("\tSubType<Double, Object> extends MyGeneric<Number> ? false\n") ;
    	System.out.println("(these results are obtained by checking compilation behaviour)") ;
    	
    	// 2. Creating SubType instances using different type parameters
    	SubType<Number, String>		sub_1 	= new SubType<Number, String>() ;    	
    	SubType<Integer, String>	sub_2 	= new SubType<Integer, String>() ;
    	SubType<Double, String>		sub_3 	= new SubType<Double, String>() ;
    	
    	// 3. Only the first casting compiles, because the 1st type parameter is the same 
    	// (if we had written "class SubType<K,V> extends MyGeneric<V> {}" it would have need to match the 2nd parameter)
    	MyGeneric<Number> 			cast_1 	= (MyGeneric<Number>) sub_1 ;
    	// MyGeneric<Number> 		cast_2 	= (MyGeneric<Number>) sub_2 ;    	
    	// MyGeneric<Number> 		cast_3 	= (MyGeneric<Number>) sub_3 ; 
	}
    
    /**
    * TYPE INFERENCE comes into play when the Java compiler can infer one or more type parameters from the source context.
    * Some practical examples are shown in the implementation of the <code>Pair</code> java class.<br/><br/>
    * 
    * Here we show some others TYPE INFERENCE examples exploiting the <code>TypeInference</code> class. 
    * 
	* To test this method run the <code>SimpleClient</code> using the <code>-g 7</code> option.<br/>
	* 
	* @see Pair
	* @see TypeInference
    */
    @SuppressWarnings({"unchecked","rawtypes"})    
    public void testTypeInference() {
    	
    	// 1. Preparing output
    	System.out.println(leadingName + "Testing type inference\n") ;    	
    	
    	// 2. Creating a TypeInference<T> Generic with a Double type parameter 
    	TypeInference<Double> object = new TypeInference<Double>( Double.valueOf(42.00D) ) ;
    	
    	// 3. Defining type parameters for methods invocation
    	Double 			double_in 	= Double.valueOf(3.1415926D) ;
    	String 			string_in	= "Hello World!" ;
    	
    	// 4.1 NO TYPE INFERENCE: we explicitly define all involved types 
    	List<String> list_1 = object.<String>genericMethod(double_in, string_in) ;
    	// 4.2 SIMPLE TYPE INFERENCE: no need to specify the method's return type, since it's inferred from context    	
    	List<String> list_2 = object.genericMethod(double_in, string_in) ;
    	// 4.3 COMPLETE TYPE INFERENCE: no need to specify any type, since it's inferred from 2nd argument
    	// Remark: this probably needs Java SE 8 or greater to avoid a "rawtypes" warning (TO BE CONFIRMED)
    	List		 list_3 = object.genericMethod(double_in, string_in) ;

    	// 5.1 Invoking a "typed argument" methods: this shows actual benefits of the above type inference 
    	object.typedMethod(list_1) ;
    	object.typedMethod(list_2) ;
    	object.typedMethod(list_3) ;
    	
    	// 6.1 Invoking a "typed argument" methods without exploiting type inference.
    	object.typedMethod( Collections.<String>emptyList() ) ;
    	// This requires Java SE 8 or greater to compile (ADVANCED TYPE INFERENCE)
    	// object.typedMethod( Collections.emptyList() ) ;
    }
    
// *********************************************************************************************************
// **********************************		INNER CLASSES 		********************************************
// *********************************************************************************************************

/** Simple class used to test the Multiple Bounds feature (setting constraints on generic methods).			*/
class ParentOne 					{	public void methodOne() { } ;	} 

/** Simple interface used to test the Multiple Bounds feature (setting constraints on generic methods).		*/
interface ParentTwo 				{	public void methodTwo() ;		}

/** Simple class used to test the Multiple Bounds feature (setting constraints on generic methods).			*/
class TestMultiple extends ParentOne implements ParentTwo {
	
	public void methodTwo() { } ;	
}

/** Simple class used to test the inheritance of generic types.												*/
class SubType<K,V> extends MyGeneric<K> {} ;

} // End of class