/**
* Title: 		 MyWildcards
* Description:	 Class testing some advanced Java 5 "Generic" Wildcards features
* Copyright:     Copyright (c) 2015
* Company:       Morpheus Quality Technologies
* Creation date: 30-12-2014
* Last modified: 20-02-2015
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.tutorial.jfive.generic ;

import java.util.List ;
import java.util.Arrays ;

/**
* This class is used by the <code>WildcardTester</code> to test invocation of method defined using Wildcard syntax and
* other "advanced" generics topics.
* These methods could have been written in the  <code>WildcardTester</code>: we chose to code them into a new class just
* to keep separated the caller object from the called instance. 
*/
public class MyWildcards {
	
  	/** Beginning of each logging message produced by this class.											*/
	protected static		String			leadingName 		= "MyWildcards - " ;
	
 // *********************************************************************************************************
 // **********************************			WILDCARDS 		********************************************
 // *********************************************************************************************************
	
    /**
    * Upper Bounded Wildcards example: when using this syntax "extends" means both "inherit" and "implements".<br/>
    * Notice also that this definition is <em>less</em> restrictive than writing <code>MyGeneric&lt;Number&gt; t</code> 
    * 
    * @param arg an instance of the MyGeneric class using a subtype of the Number type 
    */
	public void testUpperBoundedWildcards(MyGeneric<? extends Number> arg) {
		
		System.out.println() ;
		System.out.println("testUpperBoundedWildcards - Argument is : " + arg.getClass().getName()) ;
		System.out.println("testUpperBoundedWildcards - Type is     : " + arg.get().getClass().getName()) ;
		System.out.println() ;
    }
	
	/**
    * Unbounded Wildcards example: we simply use "&lt;?&gt;" as a shorthand for writing
    * <code>&lt;? extends Object&gt;</code>. 
    * This should be useful each time the method does not depend by the typed parameter. 
    * In other words, this is the correct way to (almost) use "raw types" using the Generics syntax.
    * 
    * @param arg an instance of the MyGeneric class using any general class or object   
    */
	public void testUnboundedWildcards(MyGeneric<?> arg) {
		
		System.out.println() ;
		System.out.println("testUnboundedWildcards - Argument is : " + arg.getClass().getName()) ;
		System.out.println("testUnboundedWildcards - Type is     : " + arg.get().getClass().getName()) ;
		System.out.println() ;
	}
	
	/**
    * Lower Bounded Wildcards example: this notation is basically the opposite of the "Upper Bounded Wildcards":  
    * instead of allowing (as input of the target method) only types being the children of a given class, 
    * we now allow only types being the <strong>superclass</strong> of a given class.<br/><br/>
    * 
    * In this case the input argument MUST BE a superclass of a Long type, i.e.: Object, Number or Long.
    * Notice also that this definition is <em>less</em> restrictive than writing <code>MyGeneric&lt;Long&gt; t</code> 
    * 
    * @param arg an instance of the MyGeneric class using any general class or object   
    */
	public void testLowerBoundedWildcards(MyGeneric<? super Long> arg) {
		
		System.out.println() ;
		System.out.println("testLowerBoundedWildcards - Argument is : " + arg.getClass().getName()) ;
		System.out.println("testLowerBoundedWildcards - Type is     : " + arg.get().getClass().getName()) ;
		System.out.println() ;
	}
	
 // *********************************************************************************************************
 // **********************************			CAPTURING 		********************************************
 // *********************************************************************************************************
	
	/**
    * Helper method used to show the compiler <strong>capture</strong> error.<br/><br/>
    * 
    * Such error arises when we try to use a typed argument, which has been retrieved by a Generic type, and
    * passed it again to the same Generic. Even it the type is the same, if we used a Wildcard into the 
    * method signature, the compiler can't be sure the type is really the same!
    *  
    * @param arg an instance of the MyGeneric class using any general class or object   
    */
	public void testCaptureError(MyGeneric<?> arg) {

		Object type_arg = arg.get() ;		
		System.out.println("\ntestCaptureError - got type : " + type_arg.getClass().getName()) ;
		System.out.print("\nWe can't use this argument to call set() on the Generic instance, ") ;
		System.out.println("\nbecause the Wildcard implies that the set() method can accept anything!") ;

		// This can't compile, since the produces an error like:
		// The method set(capture#15-of ?) in the type MyGeneric<capture#15-of ?> is not applicable for the arguments (Object)
		// arg.set(type_arg);
		
		// To solve the issue, we must introduce an helper method which DOES NOT use the Wildcard syntax.
		this.testCaptureErrorHelper(arg) ;
	}
	
	/**
    * Helper method used to solve the compiler capture error.<br/>
    * 
    * Notice that implementation is trivial, since we get and put the same object: this is done
    * just for training reasons.
    *  
    * @param arg an instance of the MyGeneric class using a specific typed class or object   
    */
	public <T> void testCaptureErrorHelper(MyGeneric<T> arg) {

		T type_arg = arg.get() ;		
		arg.set(type_arg);
	}
	
 // *********************************************************************************************************
 // **********************************			NON-REIFIABLE TYPES 		*********************************
 // *********************************************************************************************************
	
	/**
	* First method showing issues due to Non-Reifiable Types and Heap Pollution.<br/><br/> 
	* 
	* The method add some elements to the input typed list, using the varargs signature notation (variable arguments).<br/><br/> 
	*
	* When compiled, the varargs signature is converted to the array T[]. Then the Generics Type Erasure
	* convert it to Object[]. In some situation this could arise Heap Pollution.
    *  
    * @param list a typed list
    * @param varargs a sequence of any argument
    */
	public <T> void addElements(List<T> list, T... varargs) {
		 
		 System.out.print("\nAdding all varargs to an existing list " + list + " -> ") ;
		
		 for (T v : varargs) {
			 list.add(v) ;
			 System.out.print(" " + v) ;
		 }
		 System.out.println() ;
	 }
	 
	/**
	* Second method showing issues due to Non-Reifiable Types and Heap Pollution.<br/><br/>
	* 
	* The input varargs is converted into List<String>[], but since typed arrays are not allowed in Java,
	* actual conversion translate the input argument into List<String>[]. 
	* 
	* @param list a varargs list of <code>List&lt;String&gt;</code> elements
	*/
	@SuppressWarnings("unused")
	public void wrongMethod(List<String>... list) {
		
    	// Explicit example of varargs conversion, from List[] to an array of Objects.
    	// Since the compiler has already generated the warning (see caller):
    	// "Type safety: A generic array of List<String> is created for a varargs parameter" 
    	// -> this code does not generate any warning, even if potentially wrong.
		Object[] array = list ;
		
		// Since List extends Object, we are allowed to cast the List<Integer> into the array
		// In this way we "pollute" the heap memory, since "list" and "array" points to the same instance,
		// and now the VM believes such that object regards an Integer typed generics.
		array[0] = Arrays.asList(1) ;
		
		System.out.println("\nExample of exception due to Heap Pollution:\n") ;
		
		// If now we cast the Integer into a String, we get a ClassCastException
		String string = (String) list[0].get(0) ;
	 }
	
 // *********************************************************************************************************
 // **********************************			RESTRICITONS		 		*********************************
 // *********************************************************************************************************
	
	/**
	* Shows that we cannot instantiate directly a typed parameter (the 1st argument) and also an example
	* of workaround (the 2nd argument). This restriction is also known as "Cannot Create Instances of 
	* Type Parameters".
	* 
    * @param arg an instance of the MyGeneric class using a specific typed class or object 
    * @param type_class the class of the typed parameter 
	*/
	@SuppressWarnings("unused")
	public <T> void createTypeInstance(MyGeneric<T> arg, Class<T> type_class) {
		
			// We cannot instantiate directly the typed parameter (does not compile)
			// T a_t_object = new T() ;
		
			// Example of workaround: using Java reflection to instantiate the typed parameter
			try {
				T a_t_object = type_class.newInstance() ;
			} catch (InstantiationException e) {
				// In this scenario this workaround does not work, since "new Integer()" does not exist.
				System.out.println("\t" + leadingName + "Restriction 2: Integer does not allow an empty constructor -> 'expected exception'") ;
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			}
	}
	
	/**
	* Shows that cannot use "instanceof" at runtime on Generics type. This restriction is also known as 
	* "Cannot Use Casts or instanceof with Parameterized Types".
	* 
    * @param arg an instance of the MyGeneric class using a specific typed class or object 
	*/
	public <T> void testInstanceOf(List<T> typed, List<?> unbounded) {
		
		String localLeadingName = "\t" + leadingName + "Restriction 4: " ;  
		System.out.println(localLeadingName + "cannot use instanceof on typed parameters") ;		
		
		// This can't compile: due to Type Erasure, at runtime we can't identify the instance type  
		// if (typed instanceof List<Integer>) System.out.println("Magic!!!") ;
		
		// We can only test against an unbounded type, but such test provides no information
		if (unbounded instanceof List<?>) System.out.println(localLeadingName + "trivial test (no information)") ;
	}
	
	/**
	* Explains because we can't create arrays of generic types. This restriction is also known as 
	* "Cannot Create Arrays of Parameterized Types".
	*/
	@SuppressWarnings("rawtypes")
	public void testArray() {
		
		String localLeadingName = "\t" + leadingName + "Restriction 5: " ;  
		System.out.println(localLeadingName + "cannot create array of Generics type (see code sample for explanation)") ;
		
		// We can't create arrays of generic types: next line does not compile
		// MyGeneric<String>[] does_not_compile = new MyGeneric<String>[10] ;
		
		// Let's explain the reason. If code were compiled, the line above would become (Type Erasure):
		MyGeneric[] array = new MyGeneric[10] ;
		// Then adding elements would look like:
		array[0] = new MyGeneric<String>  ( "0" ) ; 
		array[1] = new MyGeneric<Integer> (  1  ) ;	// Triggers ArrayStoreException if array = new MyGeneric<String>[10] ;  
		array[2] = new MyGeneric<Double>  ( 2.0 ) ; // Triggers ArrayStoreException if array = new MyGeneric<String>[10] ;
	}
	
	/**
	* Shows some scenario where we can't create, catch or throw parameterized types. These scenarios are:<br/><br/>
	* 
	* 1. We can't create a Generics type extending Exception or Throwable.<br/>
	* 2. We can't catch a type parameter.<br/><br/>
	* 
	* However, we can declare that a method is going to throws exceptions of type T, as written in the signature
	* of this very method (i.e. "throws T").
	*/
	public <T extends Exception> void testExceptions() throws T {
		
		String localLeadingName = "\t" + leadingName + "Restriction 6: " ;		
		
		// 1. To verify that we can't extend Exception, try to un-comment definition of TypeExceptions (below)
		
		// 2. Below we verify that we can't catch a type parameter (even if defined in the method signature)
		try {	
			System.out.println(localLeadingName + "cannot catch a type parameter") ; 
		} 
		// Next line can't compile
		// catch(T t) { }
		catch(Exception e) { 
			System.out.println(localLeadingName + " SHOULD NEVER BE THROWN") ;
		}
	}
	
	/**
	* Shows that we can't overload a method using the same Generics signature, but with a different type
	* parameter, because after Type Erasure the two methods would have the same signature.
	*
    * @param arg an instance of the MyGeneric class using a specific typed class or object
	*/
	public <T> void testOverload(MyGeneric<Integer> arg) { 
		
		System.out.println("\t" + leadingName + "Restriction 7: doing nothing") ;
	}

	// Can't compile, since after Type Erasure would have the same signature of the method above
	// public <T> void testOverload(MyGeneric<String> arg) {}
	 
 // *********************************************************************************************************
 // **********************************		INNER CLASSES 		*********************************************
 // *********************************************************************************************************
	
    /**   
    * Dummy class used to check that we can't extend Exception.
    */
    // class TypeExceptions<T> extends Exception {} ;
	
} // End of class