/**
* Title: 		 Pair<T>
* Description:	 Class testing some advanced Java 5 "Generic" features
* Copyright:     Copyright (c) 2013-2015
* Company:       Morpheus Quality Technologies
* Creation date: 30-12-2014
* Last modified: 10-01-2015
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.tutorial.jfive.generic ;

import adriani.tutorial.jfive.generic.GenericTester.ParentOne ;
import adriani.tutorial.jfive.generic.GenericTester.ParentTwo ;


/**
* An example of "advanced" Generic class, wrapping more than one type, and showing 
* some Generic advanced features (methods and inheritance).<br/><br/>
*  
*  Some reminds about the Generic working paradigm:<br/><br/>
*  - MyClass&lt;A&gt; is NOT a subtype of MyClass&lt;B&gt; even if A is a subclass of B!<br/>
*  
*  Methods of this class should mainly be invoked by the <code>GenericTester</code> class.
*  
*  @see GenericTester
*/
public class Pair<K, V> {

	/** The first class field hosting the Generic type parameter 								*/
    private 		K 		myKey ;
	/** The second class field hosting the Generic type parameter 								*/    
    private 		V 		myValue ;
    
  	/** Beginning of each logging message produced by this class.											*/
	protected static		String		leadingName 		= "Pair<K, V> - " ;
    
    /**
    * Generic constructor, written using TYPE INFERENCE. 
    * Without type inference we should write:<br/><br/>
    * 
    * <code>public &lt;K,V&gt; Pair(K myKey, V myValue)</code><br/><br/>
    * 
    * we can avoid the syntax above because K,V are already specified in the class declaration.<br/>
    * If the constructor needed another type of parameters (not specified at class level), then 
    * we would need following syntax:<br/><br/>
    * 
    * <code>public &lt;T&gt; Pair(K myKey, V myValue)</code><br/><br/> 
    * 
    * @param myKey the first type to be wrapped by this class (the "key" like type)
    * @param myValue the second type to be wrapped by this class (the "value" like type)
    */
	public Pair(K myKey, V myValue) {
		
		this.myKey 		= myKey ;
		this.myValue 	= myValue ;
	}
	
// *********************************************************************************************************
// **********************************		SETTERS/GETTERS 	********************************************
// *********************************************************************************************************
	
	/** Setter method. 														*/
    public void setKey(K key) 		{ this.myKey 	= key ; 				}
    
    /** Setter method. 														*/
    public void setValue(V value) 	{ this.myValue 	= value ; 				}
    
    /** Getter method. 														*/
    public K getKey()   			{ return myKey ; 						}
    
    /** Getter method. 														*/
    public V getValue() 			{ return myValue ; 						}	

 // *********************************************************************************************************
 // **********************************		GENERIC METHODS 	********************************************
 // *********************************************************************************************************
    
    /**
    * Compare a pair of key-value.<br/>
    * Dummy method used to show the TYPE INFERENCE usage.<br/><br/>
    * 
    * <strong>Remark</strong>: the type parameters (K,V) depicts the INPUT arguments, not the output,
    * even if they are written before the "return data type" (i.e. "boolean").
    * 
    * @param p1 the first instance of a valid <code>Pair&lt;K,V&gt;</code> object 
    * @param p2 the second instance of a valid <code>Pair&lt;K,V&gt;</code> object
    * @return <code>true</code> if the objects wrapped by the input types are equal (same content) 
    */
	public static <K, V> boolean compare(Pair<K, V> p1, Pair<K, V> p2) {

		// Note for beginners: "equals()" return true only if the objects points to the same instance, 
		// with some exceptions for String objects: String literals are stored in the "string common pool", 
		// String objects (or modified String) are instead different objects (since Strings are immutable, i.e 
		// each modification creates a new string).
		// See: http://www.ntu.edu.sg/home/ehchua/programming/java/J3d_String.html
		/*
		String a = "Hello" ;
		String b = "Hel" + "lo" ;
		System.out.println(leadingName + "String a vs b   : " + System.identityHashCode(a) + " - " + System.identityHashCode(b)) ;
		System.out.println(leadingName + "String a == b   : " + a.equals(b)) ;
		String aa = new String("Hello") ;
		String bb = new String("Hello") ;
		System.out.println(leadingName + "String aa vs bb : " + System.identityHashCode(aa) + " - " + System.identityHashCode(bb)) ;
		System.out.println(leadingName + "String aa == bb : " + aa.equals(bb)) ;
		*/
		
		boolean keys 	= p1.getKey().equals(p2.getKey()) ;
		boolean values 	= p1.getValue().equals(p2.getValue()) ;
		
		return (keys && values) ;
	}
	
    /**
    * Inspect type parameter (T) and type argument (t).<br/>
    * Actually this is just a dummy method used to show the BOUNDED TYPE PARAMETERS usage.<br/><br/>  
    * 
    * <strong>Remark</strong>: here the keyword "extends" means both "inherits" and/or "implements".<br/><br/>
    * 
    * @param t the input type parameter, which must extend the <code>CharSequence</code> interface 
    */
	public <T extends CharSequence> void inspect(T t){
		
		// Inspection
        System.out.println("Bounded - Key      (K)                    : " + myKey.getClass().getName()		) ;
        System.out.println("Bounded - Value    (V)                    : " + myValue.getClass().getName()	) ;
        System.out.println("Bounded - Argument (t)                    : " + t.getClass().getName()			) ;
        // Since the compiler knows the data type, we can invoke CharSequence methods!
        System.out.println("Bounded - Inspect input char length   (t) : " + t.length()						) ;
        System.out.println() ;
    }	
	
    /**
    * Dummy method used to show the MULTIPLE BOUNDED TYPE PARAMETERS usage.<br/><br/>  
    * 
    * Allowed pattern: T extends CLASS & INTERFACE_1 & INTERFACE_2 etc.<br/>
    * (i.e. one Class, always first, and many Interfaces)<br/><br/>
    * 
    * Such definition implies that input type argument must match ALL CONDITIONS!
    * 
    * @param t the input type parameter, which must extend both <code>ParentOne</code> and <code>ParentTwo</code>
    */
	public <T extends ParentOne & ParentTwo> void multiple(T t){
		
        System.out.println("\nTesting  multiple(T t) method - Input class name: " + t.getClass().getName()) ;
        System.out.println() ;
    }
	
    /**
    * Example of implementation of GENERIC ALGORITHMS, using the BOUNDED TYPE PARAMETERS. 
    * This method works on any Generic type. In this implementation the only constraint is that the same input 
    * Generic type must be used both as element, both as array of elements, since the algorithm will compare
    * all array's elements with the second argument.<br/><br/>
    * 
    * This method counts the number of array elements bigger than the second argument.<br/><br/> 
    * 
    * Remark: in this example, if we did not specify that <code>T extends Comparable&lt;T&gt;</code>, 
    * it would not be possible to compare the elements.<br/><br/>
    * 
    * @param anArray an array containing element of type T  
    * @param elem a single instance of a type T
    * @return the number of array elements bigger than the second argument
    */
	public static <T extends Comparable<T>> int compare(T[] anArray, T elem) {
	// Remark: the signature below would work anyway, my generates a compile warning!
	// public static <T extends Comparable> int compare(T[] anArray, T elem) {
		
		int count = 0 ;
		for (T e : anArray) {
		    if (e.compareTo(elem) > 0) ++count ;
		}
		return count ;
	}    	
	
} // End of class