/**
* Title: 		 WildcardTester
* Description:	 Class testing some advanced Java 5 "Generic" features
* Copyright:     Copyright (c) 2013-2015
* Company:       Morpheus Quality Technologies
* Creation date: 30-12-2014
* Last modified: 20-02-2015
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.tutorial.jfive.generic ;

import java.util.List ;
import java.util.ArrayList;

import static java.util.Arrays.asList ;			// Importing a method, see Java 5 "import static" feature

/**
* When testing inheritance (see class <code>GenericTester</code>) we introduced three types 
* of bounded types:<br/><br/>
*  
* - Single Bounded Type Parameters (e.g. &lt;T extends CharSequence&gt;) see basic example 3<br/>
* - Multiple Bounded Type Parameters (e.g. &lt;T extends ParentOne & ParentTwo&gt;) see basic example 4<br/>
* - Nested Bounded Type Parameters (e.g. &lt;T extends Comparable&lt;T&gt;&gt;) see basic example 5<br/><br/>
*
* We also saw that it's not possible to extend a Generic type when inheritance is between the typed parameters 
* (see example 6 of <code>GenericTester</code> class).<br/><br/>
* 
* This class shows some example of usage of the <strong>Wildcards</strong> syntax, which allows to relax the constraint
* above. Methods of this class can be invoked by command line using the <code>-w</code> option.<br/>
* For more details see:<br/><br/> 
* 
* @see <a href="http://docs.oracle.com/javase/tutorial/java/generics/wildcards.html">wildcards (Oracle tutorial)</a>
*/
public class WildcardTester {
	
  	/** An helper class implementing some test testing wildcards											*/
	protected 				MyWildcards		tester		 		= new MyWildcards() ;

  	/** Beginning of each logging message produced by this class.											*/
	protected static		String			leadingName 		= "WildcardTester - " ;
	
    /**
	* To invoke a single test method specify one these input <code>code</code> integer parameters:<br/><br/>
	*
	* <strong>0</strong>: Test upper bounded wildcars for method signatures (e.g. A<? extends B> arg).<br/>
	* <strong>1</strong>: Test unbounded wildcars for method signatures (e.g. A<?> arg).<br/>
	* <strong>2</strong>: Test lower bounded wildcars for method signatures (e.g. A<? super B> arg).<br/>
	* <strong>3</strong>: Shows how to use wildcars to allow Generics inheritance (see 2.6).<br/>
	* <strong>4</strong>: An example of error due to the Wildcard Capture feature.<br/>
	* <strong>5</strong>: Discuss the Type Erasure mechanism (typed parameters are "replaced" when compiled).<br/>
	* <strong>6</strong>: Shows an example of possible issues arising from mixing Generics and Varargs (see 1.7).<br/>
	* <strong>7</strong>: A summary of some restrictions about usage of Generics.<br/>
	*
	* @param code the integer index used to select the Java 5 wildcards feature
    */
    public void testWildcards(int code) {
    	
    	switch(code) {
			case 	0	: upperBoundedWildcards() ;										break ;
			case 	1	: unboundedWildcards() ;										break ;
			case 	2	: lowerBoundedWildcards() ;										break ;
			case 	3	: subtypes() ;													break ;
			case 	4	: capturing() ;													break ;
			case 	5	: erase_generics() ;											break ;
			case 	6	: non_reifiable_types() ;										break ;
			case 	7	: restrictions() ;												break ;
			default		: System.out.println("Please specify a number between 0 and 7") ;
		}
	}
    
    /**
	* Basic example of usage of the Upper Bounded Wildcards.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-w 0</code> option.<br/> 
	* 
	* @see <a href="http://docs.oracle.com/javase/tutorial/java/generics/upperBounded.html">wildcards (Oracle tutorial)</a>
    */
    public void upperBoundedWildcards() {
    	
    	// 1. Creating typed Generics 
    	MyGeneric<Integer> 	generic_int = new MyGeneric<Integer>() ;
    	MyGeneric<Double> 	generic_dou = new MyGeneric<Double>() ;
    	MyGeneric<String> 	generic_str = new MyGeneric<String>() ;
    	
    	// 2. Setting typed properties    	
    	generic_int.set( Integer.valueOf(42) ) ;
    	generic_dou.set( Double.valueOf(17D) ) ;
    	generic_str.set( "This can't work" ) ;
    	
    	// 3. Testing
    	tester.testUpperBoundedWildcards(generic_int) ;
    	tester.testUpperBoundedWildcards(generic_dou) ;
//    	tester.testUpperBoundedWildcards(generic_str) ;		// This can't compile
    }
    
    /**
	* Basic example of usage of the Unbounded Wildcards.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-w 1</code> option.<br/> 
	* 
	* @see <a href="http://docs.oracle.com/javase/tutorial/java/generics/unboundedWildcards.html">wildcards (Oracle tutorial)</a>
    */
    public void unboundedWildcards() {
    	
    	// 1. Creating typed Generics: now we can use any type we like! 
    	MyGeneric<Integer> 	generic_int = new MyGeneric<Integer>() ;
    	MyGeneric<List<?>> 	generic_lis = new MyGeneric<List<?>>() ;
    	MyGeneric<String> 	generic_str = new MyGeneric<String>() ;
    	
    	// 2. Verifying we can use also the String type (see previous example)
    	generic_str.set( "Now this do work" ) ;
    	tester.testUnboundedWildcards(generic_str) ;
    	
    	System.out.println ("\tMyGeneric<Integer> extends MyGeneric<?> : " + (generic_int instanceof MyGeneric<?>)) ;    	
    	System.out.println ("\tMyGeneric<List<?>> extends MyGeneric<?> : " + (generic_lis instanceof MyGeneric<?>)) ;    	
    	System.out.println ("\tMyGeneric<String> extends MyGeneric<?>  : " + (generic_str instanceof MyGeneric<?>)) ;    	
    }
    
    /**
	* Basic example of usage of Lower Bounded Wildcards.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-w 2</code> option.<br/> 
	* 
	* @see <a href="http://docs.oracle.com/javase/tutorial/java/generics/lowerBounded.html">wildcards (Oracle tutorial)</a>
    */
    public void lowerBoundedWildcards() {
    	
    	// 1. Creating typed Generics 
    	MyGeneric<Long> 	generic_lon = new MyGeneric<Long>() ;
    	MyGeneric<Number> 	generic_num = new MyGeneric<Number>() ;
    	MyGeneric<Double> 	generic_dou = new MyGeneric<Double>() ;    	
    	
    	// 2. Setting typed properties    	
    	generic_lon.set( Long.valueOf(42) ) ;
    	// Casting a Short type into a Number type: we can do this since we wrote MyGeneric<Number>,
    	// and Short extends Number (this is basic java notion, does not regard current topic) 
    	generic_num.set( Short.valueOf("99") ) ;	 
    	generic_dou.set( Double.valueOf(17D) ) ;    	
    	
    	// 3. Testing
    	tester.testLowerBoundedWildcards(generic_lon) ;
    	tester.testLowerBoundedWildcards(generic_num) ;
    	// tester.testLowerBoundedWildcards(generic_dou) ;		// This can't compile
    }
    
    /**
	* This example shows the correct solution to the issue introduced in the example 6 of the 
	* <code>GenericTester</code>, regarding inheritance between Generics types.
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-w 3</code> option.<br/> 
	* 
	* @see <a href="http://docs.oracle.com/javase/tutorial/java/generics/subtyping.html">wildcards (Oracle tutorial)</a>
    */
    @SuppressWarnings("unused")    
    public void subtypes() {
    	
    	// 1. Providing some feedback
    	System.out.println("\nTesting subtyping.\n") ;
    	System.out.println("In sample 6th of GenericTester (option -g 6) we saw that:\n") ;
    	System.out.println("\tSubType<Integer, String> extends MyGeneric<Number> is FALSE\n") ;
    	System.out.println("even if SubType extends MyGeneric, and Integer extends Number.\n") ;
    	
    	// 2. Creating SubType instances using different type parameters
    	SubType<Integer, String>			subtype_1 	= new SubType<Integer, String>() ;
    	SubType<? extends Integer, String>	subtype_2 	= new SubType<Integer, String>() ;
    	
    	// 3. Casting back to parents 
    	// MyGeneric<Number>	wrong_1			= subtype_1 ; 		// Just recalling example 6th of GenericTester
    	// MyGeneric<Number>	wrong_2			= subtype_2 ;		// Just recalling example 6th of GenericTester
    	MyGeneric<? extends Number>	casted_1 	= subtype_1 ; 
    	MyGeneric<? extends Number>	casted_2 	= subtype_2 ; 
    	
    	System.out.println("SOLUTION: using wildcards we allow inheritance between different generics, for example:\n") ;
    	System.out.println ("\tSubType<Integer, String> extends MyGeneric<? extends Number> ") ;
    	System.out.println ("\tSubType<? extends Integer, String> extends MyGeneric<? extends Number>") ;
    	System.out.println ("\tetc...") ;
    }
    
    /**
	* This example shows how to read compile errors due to the Wildcard Capture feature.
	* "Capturing" is a specific Type Inference, i.e. a mechanism used by the compiler in order to understand
	* (at compile time) which is the nature of the type arguments written using the "?" symbol.<br/><br/>
	* 
	* See code implementation for details.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-w 4</code> option.<br/> 
	* 
	* @see <a href="http://docs.oracle.com/javase/tutorial/java/generics/capture.html">wildcards (Oracle tutorial)</a>
    */
    public void capturing() {

    	System.out.println ("\nTesting the Wildcard Capture compile error") ;
    	
    	// 1. Setting a typed parameter to the target object.
    	MyGeneric<Integer> 	generic_int = new MyGeneric<Integer>() ;
    	generic_int.set( Integer.valueOf(42) ) ;
    	
    	// 2. Tries to get and "set back" the type object into the Generic: can't compile and triggers
    	// the capture error. See implementation for details.
    	tester.testCaptureError(generic_int) ;
    }
    
    /**
	* This example explains the "Type Erasure" feature, regarding what happens "under the hood".<br/>
	* When Generics are compiled, all reference to the typed parameters are replaced with ordinary plain 
	* references to Object types (if unbounded) or a reference to the bounded class (if bounded).<br/>
	* Moreover, the compiler add some "bridge methods", i.e. some kind of "delegates" methods, having 
	* a similar signature, but allowing to invoke the typed methods passing ordinary Java objects instead
	* of typed method.<br/><br/>
	*  
	* For example, if we create a Generic using the type T as an Integer object, we would have:<br/><br/>
	* 
	* <code>public void codedMehtod(T a, T b) { ... }</code><br/>
	* <code>public void compiledMehtod(Integer a, Integer b) { ... }</code><br/>
	* <code>public void bridgedMehtod(Object a, Object b) { ... }</code><br/><br/>
	* 
	* where the <code>bridedMehtod</code> CAN NOT be invoked directly, but it's used by the JRE in order
	* to allow creation of a Generics instance using a different typed parameter, or a raw type.<br/><br/>
	* 
	* <strong>Remark: </strong> we need to know about <em>bridge methods</em> just because they could be
	* named in the stack trace exception description.<br/><br/>
	*  
	* To test this method run the <code>SimpleClient</code> using the <code>-w 5</code> option.<br/> 
	* 
	* @see <a href="http://docs.oracle.com/javase/tutorial/java/generics/genTypes.html">Erasure of Generic Types (Oracle tutorial)</a>
    */
    // Comment next annotation to see the warnings below
    @SuppressWarnings({"rawtypes", "unchecked", "unused"})
    public void erase_generics() {
    	
    	System.out.println("\nExample of exception due to the Type Erasure mechanism:\n") ;
    
    	// This code shows an error due to Type Erasure and "bad coding": since we ignore warnings,
    	// and we use Generics classes as raw type, the Type Erasure mechanism allows us to invoke the
    	// "wrong" setter method. The error will appear only later, when we try to cast the typed
    	// argument inside the wrong class.
    	// 
    	MySubGeneric mySubGeneric = new MySubGeneric() ;		// Creation of the sub generic
    	MyGeneric myGeneric = mySubGeneric ;					// Implicit casting to a raw type
    	myGeneric.set(42) ;										// Allowed because MyGeneric is a raw type 
    	String s = (String) myGeneric.myType ;  				// Will throw a Class Cast Exception at runtime !!!
    	System.out.println(leadingName + "This will never work!") ;
    	
    	// Note: to understand when "Bridge Method" are created, see implementation of the MySubGeneric class. 
    }
    
    /**
    * Non-Reifiable Types are types that can not made concrete.
    * For details about the issue discussed here, see comments in the method's implementation.<br/><br/> 
    * 
	* To test this method run the <code>SimpleClient</code> using the <code>-w 6</code> option.<br/> 
    * 
	* @see <a href="http://docs.oracle.com/javase/tutorial/java/generics/nonReifiableVarargsType.html">Non-Reifiable Types (Oracle tutorial)</a>      
    */
    // Next annotations should silence the warning, but they don't work, neither in JDK7... why? 
    // @SafeVarargs 
    // @SuppressWarnings("varargs")
    public void non_reifiable_types() {
    	
    	// 1. Using an ArrayList to create two concrete lists
    	List<String> list_1 = new ArrayList<String>( asList("a", "b", "c") ) ;
    	List<String> list_2 = new ArrayList<String>( asList("1", "2", "3") ) ;
    	
    	// 2. Using a varargs method to add a variable number of elements
    	tester.addElements(list_1, "d", "e", "f") ;
    	tester.addElements(list_2, "4", "5", "6", "7", "8", "9", "0") ;
    	
    	// 3. Now creating a "super-container" list, aimed to contain the above lists as elements
    	List< List<String> > major = new ArrayList< List<String> >() ;
    	// The compiler generates a warning, because the varargs feature converts the arguments lists
    	// into a typed array (i.e. List<String>[]), but such arrays are not allowed in Java.  
    	tester.addElements(major, list_1, list_2) ;
    	
    	// 4. Example of wrong method generation Heap pollution: see implementation
    	tester.wrongMethod( asList("Stefano"), asList("Adriani") ) ;
    }
    
    /**
    * This method lists some restrictions about using Generics.
    * For details about the restrictions discussed here, see comments in the method's implementation.<br/><br/>
    * 
	* To test this method run the <code>SimpleClient</code> using the <code>-w 7</code> option.<br/> 
    * 
	* @see <a href="http://docs.oracle.com/javase/tutorial/java/generics/restrictions.html">Restrictions on Generics (Oracle tutorial)</a>      
    */
    @SuppressWarnings("unused")
    public void restrictions() {
    	
    	System.out.println("\n" + leadingName + "Testing restrictions: see comments in the method's implementation:\n") ;
    	
    	// 1. PRIMITIVE types are not allowed as type parameter.
    	// MyGeneric<int> 	generic_int = new MyGeneric<Integer>(42) ;
    	// Solution: exploit autoboxing and autounboxing Java 5 features
    	MyGeneric<Integer> 	generic_int = new MyGeneric<Integer>(Integer.valueOf(8)) ;
    	
    	// 2. CANNOT CREATE INSTANCES of type parameters: see method implementation for details
    	tester.createTypeInstance(generic_int, Integer.class) ;
    	
    	// 3. Cannot declare STATIC FIELDS whose types are type parameters.
    	// See example in the bottom definition of this very class: the RestrictionSample<T> class
    	
    	// 4. Cannot cast or use "INSTANCEOF" with parameterized types.
    	List< List<?> > list = new ArrayList< List<?> >() ;
    	list.add (new ArrayList<String>() ) ;
    	list.add (new ArrayList<Integer>() ) ;
    	list.add (new ArrayList<Number>() ) ;
    	tester.testInstanceOf(list, list) ;
    	// 4.1 Experimenting now the "cast" restriction
    	// This can't compile, for the same reasons already seen in "option -g 6".
    	// See "GenericTester.testInheritance".
    	// List<Number> number_list = ( List<Number> ) new ArrayList<Integer>() ;
    	
    	// 5. Cannot create ARRAYS of parameterized types
    	tester.testArray() ;
    	
    	// 6. Cannot create, catch, or throw objects of parameterized types (EXCEPTION)
    	// See method implementation for details.
    	try 					{	tester.testExceptions() ;		} 
    	catch (Exception e) 	{	e.printStackTrace();			}
    	
    	// 7. Cannot OVERLOAD a method where the formal parameter types of each overload erase to the same raw type
    	// See method implementation for details.
    	tester.testOverload(generic_int) ;
    }

 // *********************************************************************************************************
 // **********************************		INNER CLASSES 		*********************************************
 // *********************************************************************************************************

    /**   
    * Simple class used to test the inheritance of generic types: it should be the same of the 
    * <code>GenericTester</code> class. It's been duplicated here just to avoid referring an inner class 
    * from another class (lazyness).
    */
    class SubType<K,V> extends MyGeneric<K> {} ;
    
    /**   
    * Simple class used to test the Type Erasure mechanism and show some examples of debug regarding    
    * the existence of the Bridge Methods.<br/>
    * Basically speaking, this class is a "string-typed children" of the <code>MyGeneric&lt;K&gt;</code>
    * Generic class.
    */
    class MySubGeneric extends MyGeneric<String> {
    	
    	/**
    	* Setter method: override the parent one.
    	*  
    	* @param arg the value of the argument (must match the class Generic type) 
    	*/
    	public void set(String s)	{
    		
    		super.set(s) ;
    	}
    	
    	// Actually, when this class is compiled, the compiler add this "hidden" method, called a BRIDGE METHOD,
    	// to ensure that the "set()" is interpreted as the override of the parent method.
    	// This because the parent class, after Type Erasure, is defined using Objects arguments.
        /*public void set(Object data) {
            set((String) data) ;
        }*/   	
    	
    } ;
    
    /**   
    * Simple class used to test some restrictions on generic types (restrictions on static fields). 
    */
    class RestrictionSample<T>  {
    	
    	/** 
    	* Example of restriction on Generics: it's not possible to define a static typed parameter: 
    	* since the field in static, and different instances could rely on different types, static
    	* typed fields are not allowed (does not compile).			
    	*/
    	// public static 		T 		my_field ;
    } ;
    
} // End of class